@@ -10,7 +10,6 @@ keithleygui configuration options.
 
 # Local import
 from keithleygui.config.user import UserConfig
-from keithley2600 import Keithley2600
 
 PACKAGE_NAME = "keithleygui"
 SUBFOLDER = ".%s" % PACKAGE_NAME
@@ -47,28 +46,34 @@ DEFAULTS = [
             "VStart": -10.0,
             "VStop": 10.0,
             "VStep": 1.0,
-            "smu_sweep": Keithley2600.SMU_LIST[0],
+            "smu_sweep": "smua",
             "tInt": 0.1,
             "pulsed": False,
             "delay": -1.0,
-            "gate": Keithley2600.SMU_LIST[0],
-            "drain": Keithley2600.SMU_LIST[1],
+            "gate": "smua",
+            "drain": "smub",
         },
     ),
-]
-
-
-for smu in Keithley2600.SMU_LIST:
-    smu_settings = (
-        smu,
+    (
+        "smua",
         {
             "sense": "SENSE_LOCAL",
             "limitv": 200.0,
             "limiti": 0.1,
             "highc": False,
         },
-    )
-    DEFAULTS.append(smu_settings)
+    ),
+    (
+        "smub",
+        {
+            "sense": "SENSE_LOCAL",
+            "limitv": 200.0,
+            "limiti": 0.1,
+            "highc": False,
+        },
+    ),
+]
+
 
 # =============================================================================
 # Config instance
