@@ -6,10 +6,14 @@
 
 # system imports
 import os.path as osp
+import configparser as cp
+
+# external imports
 import pkg_resources as pkgr
 import pyvisa
 from PyQt5 import QtCore, QtWidgets, uic
 from keithley2600 import Keithley2600, FETResultTable
+from keithley2600.keithley_driver import KeithleyIOError
 import numpy as np
 
 # local imports
@@ -20,13 +24,24 @@ from keithleygui.config.main import CONF
 MAIN_UI_PATH = pkgr.resource_filename("keithleygui", "main.ui")
 
 
+def _get_smus(keithley):
+    smu_list = [attr_name for attr_name in dir(keithley) if attr_name.startswith("smu")]
+
+    # we need at least two SMUs for GUI
+    # -> fill with placeholders if required
+    while len(smu_list) < 2:
+        smu_list.append("--")
+
+    return smu_list
+
+
 class SMUSettingsWidget(SettingsWidget):
 
     SENSE_LOCAL = 0
     SENSE_REMOTE = 1
 
     def __init__(self, smu_name):
-        SettingsWidget.__init__(self)
+        super().__init__()
 
         self.smu_name = smu_name
 
@@ -41,37 +56,43 @@ class SMUSettingsWidget(SettingsWidget):
 
     def load_defaults(self):
 
-        if CONF.get(self.smu_name, "sense") == "SENSE_LOCAL":
-            self.sense_type.setCurrentIndex(self.SENSE_LOCAL)
-        elif CONF.get(self.smu_name, "sense") == "SENSE_REMOTE":
-            self.sense_type.setCurrentIndex(self.SENSE_REMOTE)
+        if self.smu_name != "--":
 
-        self.limit_i.setValue(CONF.get(self.smu_name, "limiti"))
-        self.limit_v.setValue(CONF.get(self.smu_name, "limitv"))
-        self.high_c.setChecked(CONF.get(self.smu_name, "highc"))
+            try:
+                sense_mode = CONF.get(self.smu_name, "sense")
+
+                if sense_mode == "SENSE_LOCAL":
+                    self.sense_type.setCurrentIndex(self.SENSE_LOCAL)
+                elif sense_mode == "SENSE_REMOTE":
+                    self.sense_type.setCurrentIndex(self.SENSE_REMOTE)
+
+                self.limit_i.setValue(CONF.get(self.smu_name, "limiti"))
+                self.limit_v.setValue(CONF.get(self.smu_name, "limitv"))
+                self.high_c.setChecked(CONF.get(self.smu_name, "highc"))
+            except cp.NoSectionError:
+                pass
 
     def save_defaults(self):
 
-        if self.sense_type.currentIndex() == self.SENSE_LOCAL:
-            CONF.set(self.smu_name, "sense", "SENSE_LOCAL")
-        elif self.sense_type.currentIndex() == self.SENSE_REMOTE:
-            CONF.set(self.smu_name, "sense", "SENSE_REMOTE")
+        if self.smu_name != "--":
 
-        CONF.set(self.smu_name, "limiti", self.limit_i.value())
-        CONF.set(self.smu_name, "limitv", self.limit_v.value())
-        CONF.set(self.smu_name, "highc", self.high_c.isChecked())
+            if self.sense_type.currentIndex() == self.SENSE_LOCAL:
+                CONF.set(self.smu_name, "sense", "SENSE_LOCAL")
+            elif self.sense_type.currentIndex() == self.SENSE_REMOTE:
+                CONF.set(self.smu_name, "sense", "SENSE_REMOTE")
+
+            CONF.set(self.smu_name, "limiti", self.limit_i.value())
+            CONF.set(self.smu_name, "limitv", self.limit_v.value())
+            CONF.set(self.smu_name, "highc", self.high_c.isChecked())
 
 
 # noinspection PyArgumentList
 class SweepSettingsWidget(SettingsWidget):
     def __init__(self, keithley):
-        SettingsWidget.__init__(self)
+        super().__init__()
 
         self.keithley = keithley
-        self.smu_list = list(self.keithley.SMU_LIST)
-
-        while len(self.smu_list) < 2:
-            self.smu_list.append("--")
+        self.smu_list = _get_smus(self.keithley)
 
         self.t_int = self.addDoubleField("Integration time:", 0.1, "s", [0.000016, 0.5])
         self.t_settling = self.addDoubleField(
@@ -88,6 +109,19 @@ class SweepSettingsWidget(SettingsWidget):
         self.smu_gate.currentIndexChanged.connect(self.on_smu_gate_changed)
         self.smu_drain.currentIndexChanged.connect(self.on_smu_drain_changed)
 
+    def update_smu_list(self):
+
+        self.smu_list = _get_smus(self.keithley)
+
+        self.smu_gate.clear()
+        self.smu_drain.clear()
+
+        self.smu_gate.addItems(self.smu_list)
+        self.smu_drain.addItems(self.smu_list)
+
+        self.smu_gate.setCurrentIndex(0)
+        self.smu_drain.setCurrentIndex(1)
+
     def load_defaults(self):
 
         self.t_int.setValue(CONF.get("Sweep", "tInt"))
@@ -123,7 +157,7 @@ class SweepSettingsWidget(SettingsWidget):
 
 class TransferSweepSettingsWidget(SettingsWidget):
     def __init__(self):
-        SettingsWidget.__init__(self)
+        super().__init__()
 
         self.vg_start = self.addDoubleField("Vg start:", 0, "V")
         self.vg_stop = self.addDoubleField("Vg stop:", 0, "V")
@@ -149,7 +183,7 @@ class TransferSweepSettingsWidget(SettingsWidget):
 
 class OutputSweepSettingsWidget(SettingsWidget):
     def __init__(self):
-        SettingsWidget.__init__(self)
+        super().__init__()
 
         self.vd_start = self.addDoubleField("Vd start:", 0, "V")
         self.vd_stop = self.addDoubleField("Vd stop:", 0, "V")
@@ -173,16 +207,24 @@ class OutputSweepSettingsWidget(SettingsWidget):
 
 
 class IVSweepSettingsWidget(SettingsWidget):
-    def __init__(self, smu_list):
-        SettingsWidget.__init__(self)
+    def __init__(self, keithley):
+        super().__init__()
+        self.keithley = keithley
+        self.smu_list = _get_smus(self.keithley)
 
         self.v_start = self.addDoubleField("Vd start:", 0, "V")
         self.v_stop = self.addDoubleField("Vd stop:", 0, "V")
         self.v_step = self.addDoubleField("Vd step:", 0, "V")
-        self.smu_sweep = self.addSelectionField("Sweep SMU:", smu_list, 0)
+        self.smu_sweep = self.addSelectionField("Sweep SMU:", self.smu_list, 0)
 
         self.load_defaults()
 
+    def update_smu_list(self):
+        self.smu_list = _get_smus(self.keithley)
+        self.smu_sweep.clear()
+        self.smu_sweep.addItems(self.smu_list)
+        self.smu_sweep.setCurrentIndex(0)
+
     def load_defaults(self):
 
         self.v_start.setValue(CONF.get("Sweep", "VStart"))
@@ -215,13 +257,13 @@ class KeithleyGuiApp(QtWidgets.QMainWindow):
             lib = CONF.get("Connection", "VISA_LIBRARY")
             self.keithley = Keithley2600(address, lib)
 
-        self.smu_list = list(self.keithley.SMU_LIST)
+        self.smu_list = _get_smus(self.keithley)
         self.sweep_data = None
 
         # create sweep settings panes
         self.transfer_sweep_settings = TransferSweepSettingsWidget()
         self.output_sweep_settings = OutputSweepSettingsWidget()
-        self.iv_sweep_settings = IVSweepSettingsWidget(self.smu_list)
+        self.iv_sweep_settings = IVSweepSettingsWidget(self.keithley)
         self.general_sweep_settings = SweepSettingsWidget(self.keithley)
 
         self.tabWidgetSweeps.widget(0).layout().addWidget(self.transfer_sweep_settings)
@@ -262,6 +304,21 @@ class KeithleyGuiApp(QtWidgets.QMainWindow):
         self.connection_status_update.timeout.connect(self.update_gui_connection)
         self.connection_status_update.start(10000)  # 10 sec
 
+    def update_smu_list(self):
+
+        self.smu_list = _get_smus(self.keithley)
+
+        # recreate tabs for smu settings
+        self.smu_tabs = []
+        self.tabWidgetSettings.clear()
+        for smu_name in self.smu_list:
+            tab = SMUSettingsWidget(smu_name)
+            self.tabWidgetSettings.addTab(tab, smu_name)
+            self.smu_tabs.append(tab)
+
+        self.iv_sweep_settings.update_smu_list()
+        self.general_sweep_settings.update_smu_list()
+
     @staticmethod
     def _string_to_vd(string):
         try:
@@ -344,11 +401,8 @@ class KeithleyGuiApp(QtWidgets.QMainWindow):
         """ Start a transfer measurement with current settings."""
 
         if self.keithley.busy:
-            msg = (
-                "Keithley is currently used by another program. "
-                + "Please try again later."
-            )
-            QtWidgets.QMessageBox.information(self, str("error"), msg)
+            msg = "Keithley is currently busy. Please try again later."
+            QtWidgets.QMessageBox.information(self, "Keithley Busy", msg)
 
             return
 
@@ -404,13 +458,14 @@ class KeithleyGuiApp(QtWidgets.QMainWindow):
                 "Integration time must be between 0.001 and 25 "
                 + "power line cycles of 1/(%s Hz)." % freq
             )
-            QtWidgets.QMessageBox.information(self, str("error"), msg)
+            QtWidgets.QMessageBox.information(self, "Parameter Error", msg)
 
             return
 
         # create measurement thread with params dictionary
         self.measureThread = MeasureThread(self.keithley, params)
         self.measureThread.finished_sig.connect(self.on_measure_done)
+        self.measureThread.error_sig.connect(self.on_measure_error)
 
         # run measurement
         self._gui_state_busy()
@@ -426,12 +481,22 @@ class KeithleyGuiApp(QtWidgets.QMainWindow):
         if not self.keithley.abort_event.is_set():
             self.on_save_clicked()
 
+    def on_measure_error(self, exc):
+        self.statusBar.showMessage("    Ready.")
+        self._gui_state_idle()
+        QtWidgets.QMessageBox.information(
+            self, "Sweep Error", f"{exc.__class__.__name__}: {exc.args[0]}"
+        )
+
     @QtCore.pyqtSlot()
     def on_abort_clicked(self):
         """
         Aborts current measurement.
         """
         self.keithley.abort_event.set()
+        for smu in self.smu_list:
+            getattr(self.keithley, smu).abort()
+        self.keithley.reset()
 
     # =============================================================================
     # Interface callbacks
@@ -440,14 +505,14 @@ class KeithleyGuiApp(QtWidgets.QMainWindow):
     @QtCore.pyqtSlot()
     def on_connect_clicked(self):
         self.keithley.connect()
+        self.update_smu_list()
         self.update_gui_connection()
         if not self.keithley.connected:
             msg = (
-                "Keithley cannot be reached at %s. " % self.keithley.visa_address
-                + "Please check if address is correct and Keithley is "
-                + "turned on."
+                f"Keithley cannot be reached at {self.keithley.visa_address}. "
+                f"Please check if address is correct and Keithley is turned on."
             )
-            QtWidgets.QMessageBox.information(self, str("error"), msg)
+            QtWidgets.QMessageBox.information(self, "Connection Error", msg)
 
     @QtCore.pyqtSlot()
     def on_disconnect_clicked(self):
@@ -523,18 +588,25 @@ class KeithleyGuiApp(QtWidgets.QMainWindow):
 
     def update_gui_connection(self):
         """Check if Keithley is connected and update GUI."""
-        if self.keithley.connected and not self.keithley.busy:
+        if self.keithley.connected:
+
             try:
                 test = self.keithley.localnode.model
-                self._gui_state_idle()
-            except (pyvisa.VisaIOError, pyvisa.InvalidSession, OSError):
+            except (
+                pyvisa.VisaIOError,
+                pyvisa.InvalidSession,
+                OSError,
+                KeithleyIOError,
+            ):
                 self.keithley.disconnect()
                 self._gui_state_disconnected()
+            else:
 
-        elif self.keithley.connected and self.keithley.busy:
-            self._gui_state_busy()
-
-        elif not self.keithley.connected:
+                if self.keithley.busy:
+                    self._gui_state_busy()
+                else:
+                    self._gui_state_idle()
+        else:
             self._gui_state_disconnected()
 
     def _gui_state_busy(self):
@@ -575,8 +647,9 @@ class KeithleyGuiApp(QtWidgets.QMainWindow):
 # noinspection PyUnresolvedReferences
 class MeasureThread(QtCore.QThread):
 
-    started_sig = QtCore.Signal()
-    finished_sig = QtCore.Signal(object)
+    started_sig = QtCore.pyqtSignal()
+    finished_sig = QtCore.pyqtSignal(object)
+    error_sig = QtCore.pyqtSignal(object)
 
     def __init__(self, keithley, params):
         QtCore.QThread.__init__(self)
@@ -587,63 +660,75 @@ class MeasureThread(QtCore.QThread):
         self.wait()
 
     def run(self):
+
         self.started_sig.emit()
-        sweep_data = None
 
-        if self.params["sweep_type"] == "transfer":
-            sweep_data = self.keithley.transferMeasurement(
-                self.params["smu_gate"],
-                self.params["smu_drain"],
-                self.params["VgStart"],
-                self.params["VgStop"],
-                self.params["VgStep"],
-                self.params["VdList"],
-                self.params["tInt"],
-                self.params["delay"],
-                self.params["pulsed"],
-            )
-        elif self.params["sweep_type"] == "output":
-            sweep_data = self.keithley.outputMeasurement(
-                self.params["smu_gate"],
-                self.params["smu_drain"],
-                self.params["VdStart"],
-                self.params["VdStop"],
-                self.params["VdStep"],
-                self.params["VgList"],
-                self.params["tInt"],
-                self.params["delay"],
-                self.params["pulsed"],
-            )
+        try:
+            sweep_data = None
 
-        elif self.params["sweep_type"] == "iv":
-            direction = np.sign(self.params["VStop"] - self.params["VStart"])
-            stp = direction * abs(self.params["VStep"])
-            sweeplist = np.arange(
-                self.params["VStart"], self.params["VStop"] + stp, stp
-            )
-            v, i = self.keithley.voltageSweepSingleSMU(
-                self.params["smu_sweep"],
-                sweeplist,
-                self.params["tInt"],
-                self.params["delay"],
-                self.params["pulsed"],
-            )
+            if self.params["sweep_type"] == "transfer":
+                sweep_data = self.keithley.transfer_measurement(
+                    self.params["smu_gate"],
+                    self.params["smu_drain"],
+                    self.params["VgStart"],
+                    self.params["VgStop"],
+                    self.params["VgStep"],
+                    self.params["VdList"],
+                    self.params["tInt"],
+                    self.params["delay"],
+                    self.params["pulsed"],
+                )
+            elif self.params["sweep_type"] == "output":
+                sweep_data = self.keithley.output_measurement(
+                    self.params["smu_gate"],
+                    self.params["smu_drain"],
+                    self.params["VdStart"],
+                    self.params["VdStop"],
+                    self.params["VdStep"],
+                    self.params["VgList"],
+                    self.params["tInt"],
+                    self.params["delay"],
+                    self.params["pulsed"],
+                )
+
+            elif self.params["sweep_type"] == "iv":
+                direction = np.sign(self.params["VStop"] - self.params["VStart"])
+                stp = direction * abs(self.params["VStep"])
+
+                # forward and reverse sweeps
+                sweeplist = np.arange(
+                    self.params["VStart"], self.params["VStop"] + stp, stp
+                )
+                sweeplist = np.append(sweeplist, np.flip(sweeplist))
+
+                v, i = self.keithley.voltage_sweep_single_smu(
+                    self.params["smu_sweep"],
+                    sweeplist,
+                    self.params["tInt"],
+                    self.params["delay"],
+                    self.params["pulsed"],
+                )
+
+                params = {
+                    "sweep_type": "iv",
+                    "t_int": self.params["tInt"],
+                    "delay": self.params["delay"],
+                    "pulsed": self.params["pulsed"],
+                }
+
+                sweep_data = FETResultTable(
+                    column_titles=["Voltage", "Current"],
+                    units=["V", "A"],
+                    data=np.array([v, i]).transpose(),
+                    params=params,
+                )
 
             self.keithley.beeper.beep(0.3, 2400)
             self.keithley.reset()
 
-            params = {
-                "sweep_type": "iv",
-                "t_int": self.params["tInt"],
-                "delay": self.params["delay"],
-                "pulsed": self.params["pulsed"],
-            }
-
-            sweep_data = FETResultTable(
-                ["Voltage", "Current"], ["V", "A"], np.array([v, i]).transpose(), params
-            )
-
-        self.finished_sig.emit(sweep_data)
+            self.finished_sig.emit(sweep_data)
+        except Exception as exc:
+            self.error_sig.emit(exc)
 
 
 def run():
